from typing import List, Optional

import openrouteservice
from geopy import Point

from stpd.utils_geocode import get_point


def route_features(routes) -> dict:
    return routes['routes'][0]['summary']


class OpenRouteServiceBase:

    def __init__(
        self,
        api_key: str
    ) -> None:
        """
        Args:
            api_key: https://openrouteservice.org/
        """
        self.client = openrouteservice.Client(key=api_key)


class OpenRouteServicePathFeatures(OpenRouteServiceBase):
    """Give path receive features.
    """

    def __call__(
        self,
        location_strs: Optional[List[str]] = None,
        points: Optional[List[Point]] = None,
    ) -> dict:
        if points is not None:
            points = points
        elif location_strs is not None:
            points = [get_point(s) for s in location_strs]
        else:
            raise ValueError('At least one of location_strs or points must be provided!')
        coords = tuple([
            # this is longitude and latitude!
            # the default example is in germany, not somalia
            (p.longitude, p.latitude) for p in points
        ])
        routes = self.client.directions(coords)
        return route_features(routes)


def get_isochrones_features(
    p: Point,
    client: openrouteservice.Client,
    isochrone_seconds_range: List[float] = [100., 200., 500., 1000.],
    attributes: List[str] = ["area", "reachfactor", "total_pop"],
) -> dict:
    res = client.isochrones(
        locations=[(p.longitude, p.latitude)],
        attributes=attributes,
        range=isochrone_seconds_range,
    )
    return {
        f'range_seconds_{ft["properties"]["value"]}_{attr}': ft['properties'][attr]
        for ft in res['features']
        for attr in attributes
    }


class OpenRouteServiceLocationFeatures(OpenRouteServiceBase):
    """Give location receive features.
    """

    def __call__(
        self,
        location_str: Optional[str] = None,
        point: Optional[Point] = None,
    ) -> dict:
        if point is not None:
            point = point
        elif location_str is not None:
            point = get_point(location_str)
        else:
            raise ValueError('At least one of location_str or point must be provided!')
        return get_isochrones_features(point, client=self.client)



import logging
from datetime import datetime, timezone
from typing import Dict, List, Optional, Tuple, Union

import geopy
import geopy.distance
from geopy import Point
from OSMPythonTools.overpass import Overpass, overpassQueryBuilder

from stpd.utils_geocode import get_point

from ._osm_features import DEFAULT_FEATURE_NAMES, FEATURES

OVERPASS = Overpass()

logger = logging.getLogger(__name__)


def get_bounding_box_around(point: Point, radius_km=1.) -> Tuple[float, float, float, float]:
    center = point
    r = geopy.distance.distance(kilometers=radius_km)
    north: Point = r.destination(point=center, bearing=0)
    east: Point = r.destination(point=center, bearing=90)
    south: Point = r.destination(point=center, bearing=180)
    west: Point = r.destination(point=center, bearing=270)
    return (
        south.latitude,
        west.longitude,
        north.latitude,
        east.longitude
    )


def get_count_around(
    point: Point,
    elementType: Union[str, List[str]],
    selector: str,
    radius_km: float = 1.,
    dt: Optional[datetime] = None,
) -> int:
    query = overpassQueryBuilder(
        bbox=get_bounding_box_around(point, radius_km=radius_km),
        elementType=elementType,
        selector=selector,
        out='count'
    )
    # https://wiki.openstreetmap.org/wiki/Overpass_API/Overpass_QL#date
    p = {}
    if dt is not None:
        if dt.tzinfo is None:
            logger.warning(f'No tzinfo provided for dt:{dt}, assuming is UTC.')
        dt = dt.astimezone(timezone.utc)
        p.update({'date': f'{dt.isoformat()}'})
    result = OVERPASS.query(query, timeout=60, **p)
    return result.countElements()


class OpenStreetMap:
    """Give location receive features.
    """

    def __init__(
        self,
        feature_names: List[str] = DEFAULT_FEATURE_NAMES,
        feature_query_values: Dict[str, tuple] = {},
        radius_km=1.
    ) -> None:
        """
        Args:
            feature_names: names of features in stpd.location.OSMFeatures
            feature_query_values: dict of feature_name to (elementType, selector).
                These are passed into overpassQueryBuilder.
        """
        self.selected_features = feature_query_values
        self.selected_features.update({
            feature_name: FEATURES[feature_name]
            for feature_name in feature_names
        })
        self.radius_km = radius_km

    def __call__(
        self,
        location_str: Optional[str] = None,
        point: Optional[Point] = None,
        dt: Optional[datetime] = None,
    ) -> dict:
        if point is not None:
            point = point
        elif location_str is not None:
            point = get_point(location_str)
        else:
            raise ValueError('At least one of location_str or point must be provided!')
        res = {}
        for feature_name, v in self.selected_features.items():
            elementType, selector = v
            res[f'count_{feature_name}'] = get_count_around(
                point=point,
                elementType=elementType,
                selector=selector,
                radius_km=self.radius_km,
                dt=dt
            )
        return res



from typing import Dict, List, Tuple, Union

DEFAULT_FEATURE_NAMES = [
    'natural=tree',
    'natural=water',
    'building=yes',
    'building=house',
    'amenity=parking',
    'amenity=restaurant',
    'service=driveway',
]

FEATURES: Dict[str, Tuple[Union[str, List[str]], str]] = {
    # https://taginfo.openstreetmap.org/
    # feature_name: (elementType, selector)

    # natural:
    'natural=tree': ('node', '"natural"="tree"'),
    'natural=water': (['way', 'relation'], '"natural"="water"'),
    'natural=wood': ('relation', '"natural"="wood"'),
    'natural=scrub': ('relation', '"natural"="scrub"'),
    'natural=wetland': ('relation', '"natural"="wetland"'),

    # building:
    'building=yes': (['way', 'relation'], '"building"="yes"'),
    'building=house': ('way', '"building"="house"'),
    'building=residential': ('way', '"building"="residential"'),

    # amenity:
    'amenity=parking': (['way', 'relation'], '"amenity"="parking"'),
    'amenity=bench': ('node', '"amenity"="bench"'),
    'amenity=place_of_worship': ('node', '"amenity"="place_of_worship"'),
    'amenity=restaurant': ('node', '"amenity"="restaurant"'),
    'amenity=school': (['node', 'relation'], '"amenity"="school"'),

    # highway:
    'highway=residential': ('way', '"highway"="residential"'),
    'highway=service': ('way', '"highway"="service"'),
    'highway=track': ('way', '"highway"="track"'),
    'highway=unclassified': ('way', '"highway"="unclassified"'),
    'highway=footway': ('way', '"highway"="footway"'),
    'highway=path': ('way', '"highway"="path"'),
    'highway=tertiary': ('way', '"highway"="tertiary"'),
    'highway=crossing': ('node', '"highway"="crossing"'),
    'highway=secondary': ('way', '"highway"="secondary"'),
    'highway=primary': ('way', '"highway"="primary"'),

    # power:
    'power=tower': ('node', '"power"="tower"'),
    'power=pole': ('node', '"power"="pole"'),
    'power=generator': ('node', '"power"="generator"'),
    'power=line': ('way', '"power"="line"'),
    'power=minor_line': ('way', '"power"="minor_line"'),
    'power=substation': (['node', 'way', 'relation'], '"power"="substation"'),
    'power=portal': ('node', '"power"="portal"'),
    'power=transformer': ('node', '"power"="transformer"'),
    'power=switch': ('node', '"power"="switch"'),
    'power=catenary_mast': ('node', '"power"="catenary_mast"'),

    # service:
    'service=driveway': ('way', '"service"="driveway"'),
    'service=parking_aisle': ('way', '"service"="parking_aisle"'),
    'service=alley': ('way', '"service"="alley"'),
    'service=yard': ('way', '"service"="yard"'),
    'service=spur': ('way', '"service"="spur"'),
    'service=siding': ('way', '"service"="siding"'),
    'service=drive-through': ('way', '"service"="drive-through"'),
    'service=crossover': ('way', '"service"="crossover"'),

    # surface:
    'surface=asphalt': ('way', '"surface"="asphalt"'),
    'surface=unpaved': ('way', '"surface"="unpaved"'),
    'surface=paved': ('way', '"surface"="paved"'),
    'surface=ground': ('way', '"surface"="ground"'),
    'surface=concrete': ('way', '"surface"="concrete"'),
    'surface=gravel': ('way', '"surface"="gravel"'),
    'surface=paving_stones': ('way', '"surface"="paving_stones"'),
    'surface=dirt': ('way', '"surface"="dirt"'),
    'surface=grass': ('way', '"surface"="grass"'),
    'surface=compacted': ('way', '"surface"="compacted"'),
}



from typing import List, Tuple
from datetime import datetime, date
from holidays import country_holidays


def get_pre_post_event_dates(event_dates: List[date], dt: date) -> Tuple[date, date]:
    pre_event_dt, post_event_dt = None, None
    for event_dt in sorted(event_dates):
        if event_dt <= dt:
            pre_event_dt = event_dt
        elif pre_event_dt is not None and event_dt > dt:
            post_event_dt = event_dt
            break
    return pre_event_dt, post_event_dt


class Holiday:
    def __init__(self, country: str = 'US'):
        self.country = country

    def __call__(self, dt: datetime) -> dict:
        dt = dt.date() if isinstance(dt, datetime) else dt
        us_holidays = country_holidays('US', years=[dt.year - 1, dt.year, dt.year + 1])
        pre_event_dt, post_event_dt = get_pre_post_event_dates(
            event_dates=list(us_holidays.keys()), dt=dt)
        return {
            f'{self.country.lower()}_holiday_days_since_last': (dt - pre_event_dt).days,
            f'{self.country.lower()}_holiday_days_to_next': (post_event_dt - dt).days,
        }



from datetime import datetime

from dataclasses import dataclass
from typing import Callable, Dict

from math import cos, sin, pi, ceil


@dataclass
class Period:
    conversion_fn: Callable[[datetime], float]
    period: float


PERIODS: Dict[str, Period] = {
    'week_of_year': Period(lambda dt: dt.isocalendar()[1], period=52.286),
    'day_of_month': Period(lambda dt: dt.day, period=30.437),
    'day_of_week': Period(lambda dt: dt.weekday(), period=7.),
    'hour_of_day': Period(lambda dt: dt.hour, period=24),
    'minute_of_hour': Period(lambda dt: dt.minute, period=60),
}

REGULARITY = {
    'sine': sin,
    'cos': cos,
}


class Fourier:

    def __init__(
        self,
        week_of_year: bool = True,
        day_of_month: bool = True,
        day_of_week: bool = True,
        hour_of_day: bool = True,
        minute_of_hour: bool = True,
        use_period_fraction: float = 1.,
    ):
        self.selected_periods = {}
        if week_of_year:
            self.selected_periods['week_of_year'] = PERIODS['week_of_year']
        if day_of_month:
            self.selected_periods['day_of_month'] = PERIODS['day_of_month']
        if day_of_week:
            self.selected_periods['day_of_week'] = PERIODS['day_of_week']
        if hour_of_day:
            self.selected_periods['hour_of_day'] = PERIODS['hour_of_day']
        if minute_of_hour:
            self.selected_periods['minute_of_hour'] = PERIODS['minute_of_hour']
        self.use_period_fraction = use_period_fraction

    def __call__(self, dt: datetime) -> dict:
        out = {}
        for period_name, period in self.selected_periods.items():
            for reg_name, reg_fn in REGULARITY.items():
                for i in range(ceil(period.period * self.use_period_fraction)):
                    out[f'{period_name}_{reg_name}_phase_{i}'] = reg_fn(
                        2 * pi * (period.conversion_fn(dt) + i) / period.period
                    )
        return out



import os

from stpd.openrouteservice import (OpenRouteServiceLocationFeatures,
                                   OpenRouteServicePathFeatures)


def test_OpenRouteServicePathFeatures():
    orspathft = OpenRouteServicePathFeatures(api_key=os.getenv('ORS_SECRET'))
    orspathft(location_strs=['toronto ontario', 'hamilton ontario'])


def test_OpenRouteServiceLocationFeatures():
    orslocft = OpenRouteServiceLocationFeatures(api_key=os.getenv('ORS_SECRET'))
    orslocft(location_str='toronto ontario')



from ._openrouteservice import (OpenRouteServiceLocationFeatures,
                                OpenRouteServicePathFeatures)

__all__ = [
    'OpenRouteServiceLocationFeatures',
    'OpenRouteServicePathFeatures',
]



from geopy import Point
from OSMPythonTools.nominatim import Nominatim

nominatim = Nominatim()


def get_point(location_str: str, nom: Nominatim = nominatim) -> Point:
    best_match = nom.query(location_str).toJSON()[0]
    return Point(latitude=best_match['lat'], longitude=best_match['lon'])
